import secrets
from pathlib import Path
from io import BytesIO
import binascii

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from fastapi import FastAPI, Request
from fastapi.responses import HTMLResponse, StreamingResponse
from fastapi.templating import Jinja2Templates

here = Path(__file__).parent
data = here / "data"
app = FastAPI()

templates = Jinja2Templates(directory="templates")

encrypted = {}
MY_URL = "https://127.0.0.1:8444"


def sanitize(item: str) -> str:
    """Make sure they can't do a directory transversal attack"""
    return item.replace("\\", "").replace("/", "")


def serve_bytes_as_file(content: bytes, media_type:str = None, headers: dict = None) -> StreamingResponse:
    """FastAPI doesn't have a handy way of doing this natively so we have to improvise"""
    response = BytesIO()
    response.write(content)
    # Streaming response iterates over the data from the current seeked point
    # We need to reset it to start of file, or else it would be empty
    response.seek(0)
    return StreamingResponse(response, media_type=media_type, headers=headers)


def fix_binary_data_length(binary_data: bytes) -> bytes:
    """
    Right padding of binary data with 0 bytes
    Fix "ValueError: The length of the provided data is not a multiple of the block length."

    Taken from https://gist.github.com/btoueg/f71b62f456550da42ea9f4a4bd907d21
    """
    block_length = 16
    binary_data_length = len(binary_data)
    if binary_data_length % block_length == 0:
        return binary_data

    length_with_padding = binary_data_length + (block_length - (binary_data_length % block_length))
    return binary_data.ljust(length_with_padding, b"\0")


@app.get("/key/{enc_id}/{number}.key")
async def get_key(enc_id: str, number: int):
    if enc_id not in encrypted:
        return HTMLResponse(status_code=404)
    return serve_bytes_as_file(encrypted[enc_id][number]["key"], media_type="text/ascii")


@app.get("/playlists/{video_name}.m3u8")
async def get_playlist(video_name: str):
    playlist = here / "data" / "playlists" / f"{sanitize(video_name)}.m3u8"
    if not playlist.exists():
        return HTMLResponse(status_code=404)

    identifier = secrets.token_hex(16)  # Would be best to do check to make sure there is no conflict with existing

    m3u8_enc_line = '#EXT-X-KEY:METHOD=AES-128,URI="{base_url}/key/{identifier}/{number}.key",IV=0x{iv}'

    keys = []
    out_lines = []
    instance = -1
    for line in playlist.read_text().splitlines():
        if line.startswith("{{ video_path }}"):
            instance += 1
            # Generate a secure key and IV to use for encryption.
            key = secrets.token_bytes(16)
            iv = secrets.token_hex(16)
            keys.append({"key": key, "iv": iv})
            out_lines.insert(-1, m3u8_enc_line.format(base_url=MY_URL, identifier=identifier, number=instance, iv=iv))
            out_lines.append(line.replace("{{ video_path }}", f"/encrypted_video/{identifier}/{instance}/{video_name}"))
            continue
        out_lines.append(line)

    encrypted[identifier] = keys

    return serve_bytes_as_file("\n".join(out_lines).encode("utf-8"), media_type="application/x-mpegURL")


@app.get("/encrypted_video/{enc_id}/{number}/{video_name}/{segment_number}.ts")
async def get_segment(enc_id: str, number: int, video_name: str, segment_number: str):
    if enc_id not in encrypted:
        return HTMLResponse(status_code=404)

    if number < 0 or number > (len(encrypted[enc_id]) - 1):
        return HTMLResponse(status_code=404)

    segment = data / "video" / sanitize(video_name) / f"{segment_number}.ts"
    if not segment.exists():
        return HTMLResponse(status_code=404)

    # Encrypt the segment with the key and IV we generated during `get_playlist`
    cipher = Cipher(
        algorithm=algorithms.AES(encrypted[enc_id][number]["key"]),
        mode=modes.CBC(binascii.unhexlify(encrypted[enc_id][number]["iv"]))
    )
    encryptor = cipher.encryptor()
    ct = encryptor.update(fix_binary_data_length(segment.read_bytes())) + encryptor.finalize()

    return serve_bytes_as_file(ct, media_type="video/mp2t")


@app.get("/{video_name}.html", response_class=HTMLResponse)
async def video_template(request: Request, video_name: str):
    return templates.TemplateResponse("video.html", {"request": request, "video_name": video_name})
